@@ -5,6 +5,7 @@ use gst::prelude::*;
 use crate::signaller::{prelude::*, Signallable, Signaller};
 use crate::utils::{self, Codec, Codecs, NavigationEvent, AUDIO_CAPS, RTP_CAPS, VIDEO_CAPS};
 use crate::webrtcsrc::WebRTCSrcPad;
+use crate::RUNTIME;
 use anyhow::{Context, Error};
 use gst::glib;
 use gst::subclass::prelude::*;
@@ -12,12 +13,10 @@ use gst_webrtc::WebRTCDataChannel;
 use itertools::Itertools;
 use std::borrow::BorrowMut;
 use std::collections::{HashMap, HashSet};
-use std::mem;
 use std::str::FromStr;
 use std::sync::atomic::AtomicU16;
 use std::sync::atomic::Ordering;
-use std::sync::LazyLock;
-use std::sync::Mutex;
+use std::sync::{Arc, Condvar, LazyLock, Mutex};
 use url::Url;
 
 const DEFAULT_STUN_SERVER: Option<&str> = Some("stun://stun.l.google.com:19302");
@@ -252,7 +251,7 @@ impl ObjectImpl for BaseWebRTCSrc {
                 settings.enable_control_data_channel.to_value()
             }
             "do-retransmission" => self.settings.lock().unwrap().do_retransmission.to_value(),
-            name => panic!("{} getter not implemented", name),
+            name => panic!("{name} getter not implemented"),
         }
     }
 
@@ -335,7 +334,7 @@ struct SignallerSignals {
     handle_ice: glib::SignalHandlerId,
 }
 
-impl Session {
+impl SessionInner {
     fn webrtcbin(&self) -> gst::Bin {
         self.webrtcbin.clone().downcast::<gst::Bin>().unwrap()
     }
@@ -426,7 +425,14 @@ impl Session {
                         "Sending navigation event to peer for session {}",
                         self.id
                     );
-                    data_channel.send_string(Some(str.as_str()));
+                    if let Err(err) = data_channel.send_string_full(Some(str.as_str())) {
+                        gst::error!(
+                            CAT,
+                            obj = element,
+                            "Failed sending navigation event to peer for session {}: {err}",
+                            self.id,
+                        );
+                    }
                 }
                 None => {
                     gst::error!(
@@ -444,11 +450,12 @@ impl Session {
         &mut self,
         request: utils::ControlRequest,
         element: &super::BaseWebRTCSrc,
+        mid: Option<String>,
     ) {
         if let Some(data_channel) = &self.data_channel.borrow_mut() {
             let msg = utils::ControlRequestMessage {
                 id: self.request_counter,
-                mid: None,
+                mid,
                 request: utils::StringOrRequest::Request(request),
             };
             self.request_counter += 1;
@@ -460,7 +467,14 @@ impl Session {
                         "Sending control request to peer for session {}",
                         self.id
                     );
-                    data_channel.send_string(Some(str.as_str()));
+                    if let Err(err) = data_channel.send_string_full(Some(str.as_str())) {
+                        gst::error!(
+                            CAT,
+                            obj = element,
+                            "Failed sending control request to peer for session {}: {err}",
+                            self.id,
+                        );
+                    }
                 }
                 None => {
                     gst::error!(
@@ -549,12 +563,13 @@ impl Session {
                 move |pad, parent, buffer| {
                     let padret = gst::ProxyPad::chain_default(pad, parent, buffer);
                     let state = element.imp().state.lock().unwrap();
-                    let Some(session) = state.sessions.get(&sess_id) else {
-                        gst::error!(CAT, obj = element, "session {sess_id:?} does not exist");
-                        return padret;
-                    };
-                    let f = session.flow_combiner.lock().unwrap().update_flow(padret);
-                    f
+                    if let Some(session) = state.sessions.get(&sess_id) {
+                        let session = session.0.lock().unwrap();
+                        let f = session.flow_combiner.lock().unwrap().update_flow(padret);
+                        f
+                    } else {
+                        padret
+                    }
                 }
             ))
             .proxy_pad_event_function(glib::clone!(
@@ -570,6 +585,9 @@ impl Session {
                         let state = element.imp().state.lock().unwrap();
                         if let Some(session) = state.sessions.get(&sess_id) {
                             session
+                                .0
+                                .lock()
+                                .unwrap()
                                 .get_src_pad_from_webrtcbin_pad(&webrtcpad, &element)
                                 .map(|srcpad| {
                                     gst::event::StreamStart::builder(&srcpad.imp().stream_id())
@@ -613,30 +631,65 @@ impl Session {
                     self.id,
                     #[upgrade_or]
                     gst::PadProbeReturn::Remove,
-                    move |_pad, info| {
+                    move |pad, info| {
                         let Some(ev) = info.event() else {
                             return gst::PadProbeReturn::Ok;
                         };
-                        if ev.type_() != gst::EventType::Navigation {
-                            return gst::PadProbeReturn::Ok;
-                        };
 
-                        let mut state = element.imp().state.lock().unwrap();
-                        if let Some(session) = state.sessions.get_mut(&sess_id) {
-                            if enable_data_channel_navigation {
-                                session.send_navigation_event(
-                                    gst_video::NavigationEvent::parse(ev).unwrap(),
-                                    &element,
-                                );
+                        match ev.view() {
+                            gst::EventView::Navigation(nav) => {
+                                let mut state = element.imp().state.lock().unwrap();
+                                if let Some(session) = state.sessions.get_mut(&sess_id) {
+                                    let mut session = session.0.lock().unwrap();
+                                    if enable_data_channel_navigation {
+                                        session.send_navigation_event(
+                                            gst_video::NavigationEvent::parse(nav).unwrap(),
+                                            &element,
+                                        );
+                                    }
+                                    if enable_control_data_channel {
+                                        let transceiver = pad
+                                            .property::<gst_webrtc::WebRTCRTPTransceiver>(
+                                                "transceiver",
+                                            );
+                                        let request = utils::ControlRequest::NavigationEvent {
+                                            event: gst_video::NavigationEvent::parse(nav).unwrap(),
+                                        };
+                                        session.send_control_request(
+                                            request,
+                                            &element,
+                                            transceiver.mid().map(|s| s.to_string()),
+                                        );
+                                    }
+                                }
                             }
-                            if enable_control_data_channel {
-                                let request = utils::ControlRequest::NavigationEvent {
-                                    event: gst_video::NavigationEvent::parse(ev).unwrap(),
-                                };
-                                session.send_control_request(request, &element);
+                            gst::EventView::CustomUpstream(cup) => {
+                                let transceiver =
+                                    pad.property::<gst_webrtc::WebRTCRTPTransceiver>("transceiver");
+                                let mut state = element.imp().state.lock().unwrap();
+                                if let Some(session) = state.sessions.get_mut(&sess_id) {
+                                    let mut session = session.0.lock().unwrap();
+                                    if enable_control_data_channel {
+                                        if let Some(s) = cup.structure() {
+                                            if let Some(structure) =
+                                                utils::gvalue_to_json(&s.to_value())
+                                            {
+                                                let request =
+                                                    utils::ControlRequest::CustomUpstreamEvent {
+                                                        structure_name: s.name().to_string(),
+                                                        structure,
+                                                    };
+                                                session.send_control_request(
+                                                    request,
+                                                    &element,
+                                                    transceiver.mid().map(|s| s.to_string()),
+                                                );
+                                            }
+                                        }
+                                    }
+                                }
                             }
-                        } else {
-                            gst::error!(CAT, obj = element, "session {sess_id:?} does not exist");
+                            _ => (),
                         }
 
                         gst::PadProbeReturn::Ok
@@ -847,10 +900,26 @@ impl Session {
                         obj = element,
                         "Getting transceiver for {stream_id} and index {i} with caps: {caps:#?}"
                     );
-                    let transceiver = webrtcbin.emit_by_name::<Option<gst_webrtc::WebRTCRTPTransceiver>>(
-                        "get-transceiver",
-                        &[&(i as i32)],
-                    ).unwrap_or_else(|| {
+                    let mut transceiver = None;
+                    let mut idx = 0i32;
+                    // find the transceiver with this mline
+                    loop {
+                        let Some(to_check) = webrtcbin
+                            .emit_by_name::<Option<gst_webrtc::WebRTCRTPTransceiver>>(
+                                "get-transceiver",
+                                &[&idx],
+                            )
+                        else {
+                            break;
+                        };
+                        let mline = to_check.property::<u32>("mlineindex");
+                        if mline as usize == i {
+                            transceiver = Some(to_check);
+                            break;
+                        }
+                        idx += 1;
+                    }
+                    let transceiver = transceiver.unwrap_or_else(|| {
                         gst::warning!(CAT, "Transceiver for idx {i} does not exist, GStreamer <= 1.24, adding it ourself");
                         webrtcbin.emit_by_name::<gst_webrtc::WebRTCRTPTransceiver>(
                             "add-transceiver",
@@ -882,7 +951,7 @@ impl Session {
                     gst::error!(CAT, obj = element, "no session {session_id:?}");
                     return;
                 };
-                session.on_answer_created(reply, &element);
+                session.0.lock().unwrap().on_answer_created(reply, &element);
             }
         ));
 
@@ -911,7 +980,11 @@ impl Session {
                     return;
                 };
 
-                let (promise, webrtcbin) = session.remote_description_set(&element, &offer);
+                let (promise, webrtcbin) = session
+                    .0
+                    .lock()
+                    .unwrap()
+                    .remote_description_set(&element, &offer);
 
                 drop(state);
 
@@ -1045,30 +1118,31 @@ impl BaseWebRTCSrc {
         gst::info!(CAT, imp = self, "unpreparing");
 
         let mut state = self.state.lock().unwrap();
-        let sessions = mem::take(&mut state.sessions);
+        let session_ids: Vec<_> = state.sessions.keys().map(|k| k.to_owned()).collect();
+        let sessions: Vec<_> = session_ids
+            .iter()
+            .filter_map(|id| state.end_session(&self.obj(), id))
+            .collect();
         drop(state);
 
-        // FIXME: Needs a safer way to perform end_session.
-        // We had to release the lock during the session tear down
-        // to avoid blocking the session bin pad's chain function which could potentially
-        // block the end_session while removing the bin from webrtcsrc, causing a deadlock
-        // This looks unsafe to end a session without holding the state lock
-
-        for (_, s) in sessions.iter() {
-            let id = s.id.as_str();
-            let bin = s.webrtcbin().parent().and_downcast::<gst::Bin>().unwrap();
-            if let Err(e) = self.end_session(id, &bin) {
-                gst::error!(CAT, imp = self, "Error ending session : {e}");
-            }
+        gst::debug!(CAT, imp = self, "Ending sessions");
+        for session in sessions {
+            self.signaller().end_session(&session.0.lock().unwrap().id);
         }
-
-        // Drop all sessions after releasing the state lock. Dropping the sessions
-        // can release their bin, and during release of the bin its pads are removed
-        // but the pad-removed handler is also taking the state lock.
-        drop(sessions);
+        gst::debug!(CAT, imp = self, "All sessions have started finalizing");
 
         self.maybe_stop_signaller();
 
+        let finalizing_sessions = self.state.lock().unwrap().finalizing_sessions.clone();
+
+        let (sessions, cvar) = &*finalizing_sessions;
+        let mut sessions = sessions.lock().unwrap();
+        while !sessions.is_empty() {
+            sessions = cvar.wait(sessions).unwrap();
+        }
+
+        gst::debug!(CAT, imp = self, "All sessions are done finalizing");
+
         Ok(())
     }
 
@@ -1117,32 +1191,10 @@ impl BaseWebRTCSrc {
                     glib::closure!(
                         #[watch]
                         instance,
-                        move |_signaler: glib::Object, session_id: &str| {
+                        move |_signaller: glib::Object, session_id: &str| {
                             let this = instance.imp();
-                            let mut state = this.state.lock().unwrap();
-                            let Some(session) = state.sessions.remove(session_id) else {
-                                gst::error!(
-                                    CAT,
-                                    imp = this,
-                                    " Failed to find session {session_id}"
-                                );
-                                return false;
-                            };
 
-                            let bin = session
-                                .webrtcbin()
-                                .parent()
-                                .and_downcast::<gst::Bin>()
-                                .unwrap();
-                            drop(state);
-
-                            // FIXME: Needs a safer way to perform end_session.
-                            // We had to release the lock during the session tear down
-                            // to avoid blocking the session bin pad's chain function which could potentially
-                            // block the end_session while removing the bin from webrtcsrc, causing a deadlock
-                            // This looks unsafe to end a session without holding the state lock
-
-                            if let Err(e) = this.end_session(session_id, &bin) {
+                            if let Err(e) = this.remove_session(session_id) {
                                 gst::error!(
                                     CAT,
                                     imp = this,
@@ -1151,10 +1203,7 @@ impl BaseWebRTCSrc {
                                 return false;
                             }
 
-                            // Drop session after releasing the state lock. Dropping the session can release its bin,
-                            // and during release of the bin its pads are removed but the pad-removed handler is also
-                            // taking the state lock.
-                            drop(session);
+                            gst::log!(CAT, imp = this, "Session {session_id} cleaned up");
 
                             true
                         }
@@ -1191,7 +1240,8 @@ impl BaseWebRTCSrc {
                                 return;
                             };
 
-                            let (promise, webrtcbin) = session.handle_offer(desc, &this.obj());
+                            let (promise, webrtcbin) =
+                                session.0.lock().unwrap().handle_offer(desc, &this.obj());
                             drop(state);
                             webrtcbin
                                 .emit_by_name::<()>("set-remote-description", &[&desc, &promise]);
@@ -1219,7 +1269,7 @@ impl BaseWebRTCSrc {
                                 gst::error!(CAT, imp = this, "session {session_id:?} not found");
                                 return;
                             };
-                            session.handle_ice(
+                            session.0.lock().unwrap().handle_ice(
                                 Some(sdp_m_line_index),
                                 None,
                                 candidate,
@@ -1239,7 +1289,7 @@ impl BaseWebRTCSrc {
         &self,
         caps: &gst::Caps,
         stream_id: &str,
-        session: &mut Session,
+        session: &mut SessionInner,
     ) -> bool {
         gst::log!(CAT, "Creating pad for {caps:?}, stream: {stream_id}");
 
@@ -1333,7 +1383,7 @@ impl BaseWebRTCSrc {
         };
         drop(state);
 
-        let session = Session::new(session_id)?;
+        let session = SessionInner::new(session_id)?;
 
         let webrtcbin = session.webrtcbin();
 
@@ -1349,7 +1399,7 @@ impl BaseWebRTCSrc {
             }
         }
 
-        let bin = gst::Bin::new();
+        let bin = gst::Bin::with_name(session_id);
 
         bin.connect_pad_removed(glib::clone!(
             #[weak(rename_to = this)]
@@ -1358,11 +1408,10 @@ impl BaseWebRTCSrc {
             session_id,
             move |_, pad| {
                 let mut state = this.state.lock().unwrap();
-                let Some(session) = state.sessions.get_mut(&session_id) else {
-                    gst::warning!(CAT, imp = this, "session {session_id:?} not found");
-                    return;
+                if let Some(session) = state.sessions.get_mut(&session_id) {
+                    let session = session.0.lock().unwrap();
+                    session.flow_combiner.lock().unwrap().remove_pad(pad);
                 };
-                session.flow_combiner.lock().unwrap().remove_pad(pad);
             }
         ));
         bin.connect_pad_added(glib::clone!(
@@ -1372,11 +1421,10 @@ impl BaseWebRTCSrc {
             session_id,
             move |_, pad| {
                 let mut state = this.state.lock().unwrap();
-                let Some(session) = state.sessions.get_mut(&session_id) else {
-                    gst::warning!(CAT, imp = this, "session {session_id:?} not found");
-                    return;
+                if let Some(session) = state.sessions.get_mut(&session_id) {
+                    let session = session.0.lock().unwrap();
+                    session.flow_combiner.lock().unwrap().add_pad(pad);
                 };
-                session.flow_combiner.lock().unwrap().add_pad(pad);
             }
         ));
 
@@ -1396,7 +1444,12 @@ impl BaseWebRTCSrc {
                     gst::error!(CAT, imp = this, "session {session_id:?} not found");
                     return;
                 };
-                let bin_ghostpad = session.handle_webrtc_src_pad(&bin, pad, &this.obj());
+                let bin_ghostpad =
+                    session
+                        .0
+                        .lock()
+                        .unwrap()
+                        .handle_webrtc_src_pad(&bin, pad, &this.obj());
                 drop(state);
                 bin.add_pad(&bin_ghostpad)
                     .expect("Adding ghostpad to the bin should always work");
@@ -1410,11 +1463,10 @@ impl BaseWebRTCSrc {
             session_id,
             move |_webrtcbin, pad| {
                 let mut state = this.state.lock().unwrap();
-                let Some(session) = state.sessions.get_mut(&session_id) else {
-                    gst::error!(CAT, imp = this, "session {session_id:?} not found");
-                    return;
+                if let Some(session) = state.sessions.get_mut(&session_id) {
+                    let session = session.0.lock().unwrap();
+                    session.flow_combiner.lock().unwrap().remove_pad(pad);
                 };
-                session.flow_combiner.lock().unwrap().remove_pad(pad);
             }
         ));
 
@@ -1432,6 +1484,7 @@ impl BaseWebRTCSrc {
                         gst::error!(CAT, imp = this, "session {session_id:?} not found");
                         return;
                     };
+                    let session = session.0.lock().unwrap();
                     session.on_ice_candidate(sdp_m_line_index, candidate, &this.obj());
                 }
             ),
@@ -1452,6 +1505,7 @@ impl BaseWebRTCSrc {
                         gst::error!(CAT, imp = this, "session {session_id:?} not found");
                         return;
                     };
+                    let mut session = session.0.lock().unwrap();
                     session.on_data_channel(data_channel, &this.obj());
                 }
             ),
@@ -1465,28 +1519,69 @@ impl BaseWebRTCSrc {
             .emit_by_name::<()>("webrtcbin-ready", &[&session_id, &webrtcbin]);
 
         let mut state = self.state.lock().unwrap();
-        state.sessions.insert(session_id.to_string(), session);
+        state.sessions.insert(
+            session_id.to_string(),
+            Session(Arc::new(Mutex::new(session))),
+        );
 
         Ok(())
     }
 
-    fn end_session(&self, id: &str, bin: &gst::Bin) -> Result<(), Error> {
+    fn remove_session(&self, session_id: &str) -> Result<(), Error> {
+        let mut state = self.state.lock().unwrap();
+        if !state.sessions.contains_key(session_id) {
+            return Err(anyhow::anyhow!("No session with id: {session_id}"));
+        }
+
+        // We need to do a flush start, without which the pad chain
+        // function trying to take the state lock causes a deadlock
+        // with the state lock already acquired here.
+        self.flush_src_pad(session_id)?;
+
+        if let Some(session) = state.end_session(&self.obj(), session_id) {
+            drop(state);
+            let session = session.0.lock().unwrap();
+            self.signaller().end_session(&session.id);
+        }
+
+        self.remove_src_pad(session_id)?;
+
+        Ok(())
+    }
+
+    fn flush_src_pad(&self, session_id: &str) -> Result<(), Error> {
         let obj = self.obj();
 
-        // set the session's bin to Null and remove it
-        bin.set_state(gst::State::Null)?;
-        obj.remove(bin)?;
-
-        for pad in obj.src_pads() {
-            if pad.name().contains(id) {
-                if !pad.push_event(gst::event::Eos::new()) {
-                    gst::warning!(CAT, imp = self, "failed to send EOS on {}", pad.name());
-                }
-                obj.remove_pad(&pad)
-                    .map_err(|err| anyhow::anyhow!("Couldn't remove pad? {err:?}"))?;
+        for pad in obj
+            .src_pads()
+            .iter()
+            .filter(|pad| pad.name().contains(session_id))
+        {
+            if !pad.push_event(gst::event::FlushStart::new()) {
+                return Err(anyhow::anyhow!(
+                    "Failed to send flush-start for pad: {}",
+                    pad.name()
+                ));
             }
         }
-        // self.signaller().end_session(id);
+
+        Ok(())
+    }
+
+    fn remove_src_pad(&self, session_id: &str) -> Result<(), Error> {
+        let obj = self.obj();
+
+        for pad in obj
+            .src_pads()
+            .iter()
+            .filter(|pad| pad.name().contains(session_id))
+        {
+            obj.remove_pad(pad)
+                .map_err(|err| anyhow::anyhow!("Couldn't remove pad? {err:?}"))?;
+
+            gst::log!(CAT, imp = self, "Removed pad {}", pad.name());
+        }
+
         Ok(())
     }
 }
@@ -1599,6 +1694,9 @@ impl ElementImpl for BaseWebRTCSrc {
                             .values_mut()
                             .next()
                             .unwrap()
+                            .0
+                            .lock()
+                            .unwrap()
                             .send_navigation_event(
                                 gst_video::NavigationEvent::parse(ev).unwrap(),
                                 &self.obj(),
@@ -1614,7 +1712,10 @@ impl ElementImpl for BaseWebRTCSrc {
                             .values_mut()
                             .next()
                             .unwrap()
-                            .send_control_request(request, &self.obj());
+                            .0
+                            .lock()
+                            .unwrap()
+                            .send_control_request(request, &self.obj(), None);
                     }
                     true
                 }
@@ -1658,7 +1759,7 @@ enum SignallerState {
     Stopped,
 }
 
-struct Session {
+struct SessionInner {
     id: String,
     webrtcbin: gst::Element,
     data_channel: Option<WebRTCDataChannel>,
@@ -1668,10 +1769,15 @@ struct Session {
     request_counter: u64,
     pending_srcpads: HashMap<String, (WebRTCSrcPad, gst::Caps)>,
 }
+
+#[derive(Clone)]
+struct Session(Arc<Mutex<SessionInner>>);
+
 struct State {
     sessions: HashMap<String, Session>,
     signaller_state: SignallerState,
     signaller_signals: Option<SignallerSignals>,
+    finalizing_sessions: Arc<(Mutex<HashMap<String, Session>>, Condvar)>,
 }
 
 impl Default for State {
@@ -1680,10 +1786,65 @@ impl Default for State {
             signaller_state: SignallerState::Stopped,
             sessions: HashMap::new(),
             signaller_signals: Default::default(),
+            finalizing_sessions: Arc::new((Mutex::new(HashMap::new()), Condvar::new())),
         }
     }
 }
 
+impl State {
+    fn finalize_session(&mut self, element: &super::BaseWebRTCSrc, session: Session) {
+        let inner = session.0.lock().unwrap();
+        let session_id = inner.id.clone();
+        drop(inner);
+
+        gst::info!(CAT, "Ending session {}", session_id);
+
+        let finalizing_sessions = self.finalizing_sessions.clone();
+        let (sessions, _cvar) = &*finalizing_sessions;
+        sessions.lock().unwrap().insert(session_id.clone(), session);
+
+        let element = element.clone();
+        RUNTIME.spawn_blocking(move || {
+            let (sessions, cvar) = &*finalizing_sessions;
+            let mut sessions = sessions.lock().unwrap();
+            let session = sessions.remove(&session_id).unwrap();
+            let session = session.0.lock().unwrap();
+
+            let bin = session
+                .webrtcbin()
+                .parent()
+                .and_downcast::<gst::Bin>()
+                .unwrap();
+
+            let _ = bin.set_state(gst::State::Null);
+
+            if let Some(webrtcsrc) = bin.parent() {
+                if let Err(e) = webrtcsrc.downcast_ref::<gst::Bin>().unwrap().remove(&bin) {
+                    gst::error!(
+                        CAT,
+                        obj = bin,
+                        "Failed to remove bin for session {} :{e}",
+                        session.id
+                    );
+                }
+            }
+
+            cvar.notify_one();
+
+            gst::debug!(CAT, obj = element, "Session {session_id} ended");
+        });
+    }
+
+    fn end_session(&mut self, element: &super::BaseWebRTCSrc, session_id: &str) -> Option<Session> {
+        if let Some(session) = self.sessions.remove(session_id) {
+            self.finalize_session(element, session.clone());
+            return Some(session);
+        }
+
+        None
+    }
+}
+
 #[derive(Default)]
 pub struct WebRTCSrc {}
 
@@ -1769,7 +1930,7 @@ impl URIHandlerImpl for WebRTCSrc {
 
     fn set_uri(&self, uri: &str) -> Result<(), glib::Error> {
         let uri = Url::from_str(uri)
-            .map_err(|err| glib::Error::new(gst::URIError::BadUri, &format!("{:?}", err)))?;
+            .map_err(|err| glib::Error::new(gst::URIError::BadUri, &format!("{err:?}")))?;
 
         let socket_scheme = match uri.scheme() {
             "gstwebrtc" => Ok("ws"),
@@ -1939,7 +2100,7 @@ pub(super) mod livekit {
             .field("can-subscribe", participant_permission.can_subscribe)
             .field("can-publish", participant_permission.can_publish)
             .field("can-publish-data", participant_permission.can_publish_data)
-            .field_from_iter::<gst::Array>(
+            .field_from_iter::<gst::Array, _>(
                 "can-publish-sources",
                 participant_permission
                     .can_publish_sources
@@ -2006,7 +2167,7 @@ pub(super) mod livekit {
                     .map(|s| s.as_str_name())
                     .unwrap_or("unknown"),
             )
-            .field_from_iter::<gst::Array>(
+            .field_from_iter::<gst::Array, _>(
                 "tracks",
                 participant_info.tracks.iter().map(track_info_to_structure),
             )
@@ -2160,3 +2321,244 @@ pub(super) mod livekit {
     impl GhostPadImpl for LiveKitWebRTCSrcPad {}
     impl WebRTCSrcPadImpl for LiveKitWebRTCSrcPad {}
 }
+
+#[cfg(feature = "janus")]
+pub(super) mod janus {
+    use super::*;
+    use crate::{
+        janusvr_signaller::{JanusVRSignallerStr, JanusVRSignallerU64},
+        webrtcsink::JanusVRSignallerState,
+        webrtcsrc::WebRTCSignallerRole,
+    };
+
+    static CAT: LazyLock<gst::DebugCategory> = LazyLock::new(|| {
+        gst::DebugCategory::new(
+            "janusvrwebrtcsrc",
+            gst::DebugColorFlags::empty(),
+            Some("WebRTC Janus Video Room src"),
+        )
+    });
+
+    #[derive(Debug, Clone, Default)]
+    struct JanusSettings {
+        use_string_ids: bool,
+    }
+
+    #[derive(Debug, Default)]
+    struct JanusState {
+        janus_state: JanusVRSignallerState,
+    }
+
+    #[derive(Default, glib::Properties)]
+    #[properties(wrapper_type = crate::webrtcsrc::JanusVRWebRTCSrc)]
+    pub struct JanusVRWebRTCSrc {
+        /**
+         * GstJanusVRWebRTCSrc:use-string-ids:
+         *
+         * By default Janus uses `u64` ids to identify the room, the feed, etc.
+         * But it can be changed to strings using the `strings_ids` option in `janus.plugin.videoroom.jcfg`.
+         * In such case, `janusvrwebrtcsrc` has to be created using `use-string-ids=true` so its signaller
+         * uses the right types for such ids and properties.
+         *
+         * Since: plugins-rs-0.14.0
+         */
+        #[property(name="use-string-ids", get, construct_only, type = bool, member = use_string_ids, blurb = "Use strings instead of u64 for Janus IDs, see strings_ids config option in janus.plugin.videoroom.jcfg")]
+        settings: Mutex<JanusSettings>,
+        /**
+         * GstJanusVRWebRTCSrc:janus-state:
+         *
+         * The current state of the signaller.
+         * Since: plugins-rs-0.14.0
+         */
+        #[property(
+        name = "janus-state",
+        get,
+        member = janus_state,
+        type = JanusVRSignallerState,
+        blurb = "The current state of the signaller",
+        builder(JanusVRSignallerState::Initialized)
+    )]
+        state: Mutex<JanusState>,
+    }
+
+    #[glib::derived_properties]
+    impl ObjectImpl for JanusVRWebRTCSrc {
+        fn constructed(&self) {
+            self.parent_constructed();
+
+            let settings = self.settings.lock().unwrap();
+            let element = self.obj();
+            let ws = element
+                .upcast_ref::<crate::webrtcsrc::BaseWebRTCSrc>()
+                .imp();
+
+            let signaller: Signallable = if settings.use_string_ids {
+                JanusVRSignallerStr::new(WebRTCSignallerRole::Consumer).upcast()
+            } else {
+                JanusVRSignallerU64::new(WebRTCSignallerRole::Consumer).upcast()
+            };
+
+            let self_weak = self.downgrade();
+            signaller.connect("state-updated", false, move |args| {
+                let self_ = self_weak.upgrade()?;
+                let janus_state = args[1].get::<JanusVRSignallerState>().unwrap();
+
+                {
+                    let mut state = self_.state.lock().unwrap();
+                    state.janus_state = janus_state;
+                }
+
+                gst::debug!(
+                    CAT,
+                    imp = self_,
+                    "signaller state updated: {:?}",
+                    janus_state
+                );
+
+                self_.obj().notify("janus-state");
+
+                None
+            });
+
+            let _ = ws.set_signaller(signaller);
+        }
+    }
+
+    impl GstObjectImpl for JanusVRWebRTCSrc {}
+
+    impl ElementImpl for JanusVRWebRTCSrc {
+        fn metadata() -> Option<&'static gst::subclass::ElementMetadata> {
+            static ELEMENT_METADATA: LazyLock<gst::subclass::ElementMetadata> =
+                LazyLock::new(|| {
+                    gst::subclass::ElementMetadata::new(
+                        "JanusVRWebRTCSrc",
+                        "Source/Network/WebRTC",
+                        "WebRTC source with Janus Video Room signaller",
+                        "Eva Pace <epace@igalia.com>",
+                    )
+                });
+
+            Some(&*ELEMENT_METADATA)
+        }
+    }
+
+    impl BinImpl for JanusVRWebRTCSrc {}
+
+    impl BaseWebRTCSrcImpl for JanusVRWebRTCSrc {}
+
+    #[glib::object_subclass]
+    impl ObjectSubclass for JanusVRWebRTCSrc {
+        const NAME: &'static str = "GstJanusVRWebRTCSrc";
+        type Type = crate::webrtcsrc::JanusVRWebRTCSrc;
+        type ParentType = crate::webrtcsrc::BaseWebRTCSrc;
+        type Interfaces = (gst::URIHandler,);
+    }
+
+    impl URIHandlerImpl for JanusVRWebRTCSrc {
+        const URI_TYPE: gst::URIType = gst::URIType::Src;
+
+        fn protocols() -> &'static [&'static str] {
+            &["gstjanusvr", "gstjanusvrs"]
+        }
+
+        fn uri(&self) -> Option<String> {
+            {
+                let settings = self.settings.lock().unwrap();
+                if settings.use_string_ids {
+                    // URI not supported for string ids
+                    return None;
+                }
+            }
+
+            let obj = self.obj();
+            let base = obj.upcast_ref::<crate::webrtcsrc::BaseWebRTCSrc>().imp();
+            let signaller = base.signaller();
+
+            let janus_endpoint = signaller.property::<String>("janus-endpoint");
+            let uri = janus_endpoint
+                .replace("wss://", "gstjansvrs://")
+                .replace("ws://", "gstjanusvr://");
+            let room_id = signaller.property::<u64>("room-id");
+            let producer_peer_id = signaller.property::<u64>("producer-peer-id");
+
+            Some(format!(
+                "{uri}?use-string-ids=false&room-id={room_id}&producer-peer-id={producer_peer_id}"
+            ))
+        }
+
+        fn set_uri(&self, uri: &str) -> Result<(), glib::Error> {
+            gst::debug!(CAT, imp = self, "parsing URI {uri}");
+
+            let uri = Url::from_str(uri)
+                .map_err(|err| glib::Error::new(gst::URIError::BadUri, &format!("{err:?}")))?;
+
+            let socket_scheme = match uri.scheme() {
+                "gstjanusvr" => Ok("ws"),
+                "gstjanusvrs" => Ok("wss"),
+                _ => Err(glib::Error::new(
+                    gst::URIError::BadUri,
+                    &format!("Invalid protocol: {}", uri.scheme()),
+                )),
+            }?;
+
+            let port = uri
+                .port()
+                .map(|port| format!(":{port}"))
+                .unwrap_or_default();
+
+            let janus_endpoint = format!(
+                "{socket_scheme}://{}{port}{}",
+                uri.host_str().unwrap_or("127.0.0.1"),
+                uri.path()
+            );
+
+            let use_strings_ids = uri
+                .query_pairs()
+                .find(|(k, _v)| k == "use-string-ids")
+                .map(|(_k, v)| v.to_lowercase() == "true")
+                .unwrap_or_default();
+            if use_strings_ids {
+                // TODO: we'd have to instantiate a JanusVRSignallerStr and set it on the src element
+                // but "signaller" is a construct-only property.
+                return Err(glib::Error::new(
+                    gst::URIError::BadUri,
+                    "use-string-ids=true not yet supported in URI",
+                ));
+            }
+
+            let room_id = uri
+                .query_pairs()
+                .find(|(k, _v)| k == "room-id")
+                .map(|(_k, v)| v)
+                .ok_or(glib::Error::new(gst::URIError::BadUri, "room-id missing"))?;
+            let producer_peer_id = uri
+                .query_pairs()
+                .find(|(k, _v)| k == "producer-peer-id")
+                .map(|(_k, v)| v)
+                .ok_or(glib::Error::new(
+                    gst::URIError::BadUri,
+                    "producer-peer-id missing",
+                ))?;
+
+            let obj = self.obj();
+            let base = obj.upcast_ref::<crate::webrtcsrc::BaseWebRTCSrc>().imp();
+            let signaller = base.signaller();
+
+            let room_id = room_id.parse::<u64>().map_err(|err| {
+                glib::Error::new(gst::URIError::BadUri, &format!("Invalid room-id: {err}"))
+            })?;
+            let producer_peer_id = producer_peer_id.parse::<u64>().map_err(|err| {
+                glib::Error::new(
+                    gst::URIError::BadUri,
+                    &format!("Invalid producer-peer-id: {err}"),
+                )
+            })?;
+
+            signaller.set_property("janus-endpoint", &janus_endpoint);
+            signaller.set_property("room-id", room_id);
+            signaller.set_property("producer-peer-id", producer_peer_id);
+
+            Ok(())
+        }
+    }
+}
